@@ -9,7 +9,7 @@ import (
 	"github.com/mattn/go-runewidth"
 )
 
-func init() { //nolint:gochecknoinits
+func init() {
 	runewidth.DefaultCondition.EastAsianWidth = false
 }
 
@@ -36,8 +36,8 @@ type style struct {
 }
 
 type Colors struct {
-	Background string `json:"background"`
-	Foreground string `json:"foreground"`
+	Background string `json:"background" toml:"background"`
+	Foreground string `json:"foreground" toml:"foreground"`
 }
 
 const (
@@ -54,7 +54,7 @@ const (
 	// Foreground takes the current segment's foreground color
 	Foreground = "foreground"
 
-	AnchorRegex    = `^(?P<ANCHOR><(?P<FG>[^,>]+)?,?(?P<BG>[^>]+)?>)`
+	AnchorRegex    = `^(?P<ANCHOR><(?P<FG>[^,<>]+)?,?(?P<BG>[^<>]+)?>)`
 	colorise       = "\x1b[%sm"
 	transparent    = "\x1b[0m\x1b[%s;49m\x1b[7m"
 	transparentEnd = "\x1b[27m"
@@ -66,12 +66,14 @@ const (
 	OSC7  = "osc7"
 	OSC51 = "osc51"
 
-	LINK   = "link"
-	TEXT   = "text"
-	OTHER  = "other"
 	ANCHOR = "ANCHOR"
 	BG     = "BG"
 	FG     = "FG"
+
+	hyperLinkStart   = "<LINK>"
+	hyperLinkEnd     = "</LINK>"
+	hyperLinkText    = "<TEXT>"
+	hyperLinkTextEnd = "</TEXT>"
 )
 
 // Writer writes colorized ANSI strings
@@ -86,18 +88,17 @@ type Writer struct {
 	builder strings.Builder
 	length  int
 
-	foreground        Color
-	background        Color
-	currentForeground Color
-	currentBackground Color
-	runes             []rune
-	transparent       bool
-	invisible         bool
+	foreground  Color
+	background  Color
+	current     ColorHistory
+	runes       []rune
+	transparent bool
+	invisible   bool
+	hyperlink   bool
 
 	shell                 string
 	format                string
 	left                  string
-	right                 string
 	title                 string
 	linechange            string
 	clearBelow            string
@@ -106,28 +107,22 @@ type Writer struct {
 	restoreCursorPosition string
 	escapeLeft            string
 	escapeRight           string
-	hyperlink             string
-	hyperlinkRegex        string
 	osc99                 string
 	osc7                  string
 	osc51                 string
 
-	// hyperlink
-	hasHyperlink             bool
-	hyperlinkBuilder         strings.Builder
-	bracketIndex, roundCount int
-	hyperlinkState           string
+	hyperlinkStart  string
+	hyperlinkCenter string
+	hyperlinkEnd    string
 }
 
 func (w *Writer) Init(shellName string) {
-	w.hyperlinkState = OTHER
 	w.shell = shellName
 	w.format = "%s"
 	switch w.shell {
 	case shell.BASH:
 		w.format = "\\[%s\\]"
 		w.linechange = "\\[\x1b[%d%s\\]"
-		w.right = "\\[\x1b[%dC\\]"
 		w.left = "\\[\x1b[%dD\\]"
 		w.clearBelow = "\\[\x1b[0J\\]"
 		w.clearLine = "\\[\x1b[K\\]"
@@ -136,15 +131,15 @@ func (w *Writer) Init(shellName string) {
 		w.title = "\\[\x1b]0;%s\007\\]"
 		w.escapeLeft = "\\["
 		w.escapeRight = "\\]"
-		w.hyperlink = "\\[\x1b]8;;%s\x1b\\\\\\]%s\\[\x1b]8;;\x1b\\\\\\]"
-		w.hyperlinkRegex = `(?P<STR>\\\[\x1b\]8;;(.+)\x1b\\\\\\\](?P<TEXT>.+)\\\[\x1b\]8;;\x1b\\\\\\\])`
+		w.hyperlinkStart = "\\[\x1b]8;;"
+		w.hyperlinkCenter = "\x1b\\\\\\]"
+		w.hyperlinkEnd = "\\[\x1b]8;;\x1b\\\\\\]"
 		w.osc99 = "\\[\x1b]9;9;%s\x1b\\\\\\]"
 		w.osc7 = "\\[\x1b]7;file://%s/%s\x1b\\\\\\]"
 		w.osc51 = "\\[\x1b]51;A;%s@%s:%s\x1b\\\\\\]"
 	case shell.ZSH, shell.TCSH:
 		w.format = "%%{%s%%}"
 		w.linechange = "%%{\x1b[%d%s%%}"
-		w.right = "%%{\x1b[%dC%%}"
 		w.left = "%%{\x1b[%dD%%}"
 		w.clearBelow = "%{\x1b[0J%}"
 		w.clearLine = "%{\x1b[K%}"
@@ -153,14 +148,14 @@ func (w *Writer) Init(shellName string) {
 		w.title = "%%{\x1b]0;%s\007%%}"
 		w.escapeLeft = "%{"
 		w.escapeRight = "%}"
-		w.hyperlink = "%%{\x1b]8;;%s\x1b\\%%}%s%%{\x1b]8;;\x1b\\%%}"
-		w.hyperlinkRegex = `(?P<STR>%{\x1b]8;;(.+)\x1b\\%}(?P<TEXT>.+)%{\x1b]8;;\x1b\\%})`
+		w.hyperlinkStart = "%{\x1b]8;;"
+		w.hyperlinkCenter = "\x1b\\%}"
+		w.hyperlinkEnd = "%{\x1b]8;;\x1b\\%}"
 		w.osc99 = "%%{\x1b]9;9;%s\x1b\\%%}"
 		w.osc7 = "%%{\x1b]7;file://%s/%s\x1b\\%%}"
 		w.osc51 = "%%{\x1b]51;A%s@%s:%s\x1b\\%%}"
 	default:
 		w.linechange = "\x1b[%d%s"
-		w.right = "\x1b[%dC"
 		w.left = "\x1b[%dD"
 		w.clearBelow = "\x1b[0J"
 		w.clearLine = "\x1b[K"
@@ -170,8 +165,9 @@ func (w *Writer) Init(shellName string) {
 		// when in fish on Linux, it seems hyperlinks ending with \\ print a \
 		// unlike on macOS. However, this is a fish bug, so do not try to fix it here:
 		// https://github.com/JanDeDobbeleer/oh-my-posh/pull/3288#issuecomment-1369137068
-		w.hyperlink = "\x1b]8;;%s\x1b\\%s\x1b]8;;\x1b\\"
-		w.hyperlinkRegex = "(?P<STR>\x1b]8;;(.+)\x1b\\\\\\\\?(?P<TEXT>.+)\x1b]8;;\x1b\\\\)"
+		w.hyperlinkStart = "\x1b]8;;"
+		w.hyperlinkCenter = "\x1b\\"
+		w.hyperlinkEnd = "\x1b]8;;\x1b\\"
 		w.osc99 = "\x1b]9;9;%s\x1b\\"
 		w.osc7 = "\x1b]7;file://%s/%s\x1b\\"
 		w.osc51 = "\x1b]51;A%s@%s:%s\x1b\\"
@@ -195,15 +191,6 @@ func (w *Writer) SetParentColors(background, foreground string) {
 	}}, w.ParentColors...)
 }
 
-func (w *Writer) CarriageForward() string {
-	return fmt.Sprintf(w.right, 1000)
-}
-
-func (w *Writer) GetCursorForRightWrite(length, offset int) string {
-	strippedLen := length + (-offset)
-	return fmt.Sprintf(w.left, strippedLen)
-}
-
 func (w *Writer) ChangeLine(numberOfLines int) string {
 	if w.Plain {
 		return ""
@@ -306,7 +293,7 @@ func (w *Writer) Write(background, foreground, text string) {
 	}
 	// validate if we start with a color override
 	match := regex.FindNamedRegexMatch(AnchorRegex, text)
-	if len(match) != 0 {
+	if len(match) != 0 && match[ANCHOR] != hyperLinkStart {
 		colorOverride := true
 		for _, style := range knownStyles {
 			if match[ANCHOR] != style.AnchorStart {
@@ -315,23 +302,29 @@ func (w *Writer) Write(background, foreground, text string) {
 			w.writeEscapedAnsiString(style.Start)
 			colorOverride = false
 		}
+
 		if colorOverride {
-			w.currentBackground, w.currentForeground = w.asAnsiColors(match[BG], match[FG])
+			w.current.Add(w.asAnsiColors(match[BG], match[FG]))
 		}
 	}
+
 	w.writeSegmentColors()
 
+	// print the hyperlink part AFTER the coloring
+	if match[ANCHOR] == hyperLinkStart {
+		w.hyperlink = true
+		w.builder.WriteString(w.hyperlinkStart)
+	}
+
 	text = text[len(match[ANCHOR]):]
 	w.runes = []rune(text)
-
-	// only run hyperlink logic when we have to
-	w.hasHyperlink = strings.Count(text, "«")+strings.Count(text, "»")+strings.Count(text, "(")+strings.Count(text, ")") >= 4
+	hyperlinkTextPosition := 0
 
 	for i := 0; i < len(w.runes); i++ {
 		s := w.runes[i]
 		// ignore everything which isn't overriding
 		if s != '<' {
-			w.write(i, s)
+			w.write(s)
 			continue
 		}
 
@@ -339,25 +332,46 @@ func (w *Writer) Write(background, foreground, text string) {
 		text = string(w.runes[i:])
 		match = regex.FindNamedRegexMatch(AnchorRegex, text)
 		if len(match) > 0 {
-			i = w.writeColorOverrides(match, background, i)
+			// check for hyperlinks first
+			switch match[ANCHOR] {
+			case hyperLinkStart:
+				w.hyperlink = true
+				i += len([]rune(match[ANCHOR])) - 1
+				w.builder.WriteString(w.hyperlinkStart)
+				continue
+			case hyperLinkText:
+				w.hyperlink = false
+				i += len([]rune(match[ANCHOR])) - 1
+				hyperlinkTextPosition = i
+				w.builder.WriteString(w.hyperlinkCenter)
+				continue
+			case hyperLinkTextEnd:
+				// this implies there's no text in the hyperlink
+				if hyperlinkTextPosition+1 == i {
+					w.builder.WriteString("link")
+					w.length += 4
+				}
+				i += len([]rune(match[ANCHOR])) - 1
+				continue
+			case hyperLinkEnd:
+				i += len([]rune(match[ANCHOR])) - 1
+				w.builder.WriteString(w.hyperlinkEnd)
+				continue
+			}
+
+			i = w.writeArchorOverride(match, background, i)
 			continue
 		}
 
 		w.length += runewidth.RuneWidth(s)
-		w.write(i, s)
+		w.write(s)
 	}
 
-	// append remnant hyperlink
-	w.builder.WriteString(w.hyperlinkBuilder.String())
-	w.hyperlinkBuilder.Reset()
-	w.hyperlinkState = OTHER
-
 	// reset colors
 	w.writeEscapedAnsiString(resetStyle.End)
 
-	// reset current
-	w.currentBackground = ""
-	w.currentForeground = ""
+	// pop last color from the stack
+	w.current.Pop()
 }
 
 func (w *Writer) String() (string, int) {
@@ -373,29 +387,39 @@ func (w *Writer) writeEscapedAnsiString(text string) {
 	if w.Plain {
 		return
 	}
+
 	if len(w.format) != 0 {
 		text = fmt.Sprintf(w.format, text)
 	}
-	if w.hyperlinkState == OTHER {
-		w.builder.WriteString(text)
-		return
-	}
-	w.hyperlinkBuilder.WriteString(text)
+
+	w.builder.WriteString(text)
 }
 
 func (w *Writer) getAnsiFromColorString(colorString string, isBackground bool) Color {
 	return w.AnsiColors.ToColor(colorString, isBackground, w.TrueColor)
 }
 
+func (w *Writer) write(s rune) {
+	if w.invisible {
+		return
+	}
+
+	if !w.hyperlink {
+		w.length += runewidth.RuneWidth(s)
+	}
+
+	w.builder.WriteRune(s)
+}
+
 func (w *Writer) writeSegmentColors() {
 	// use correct starting colors
 	bg := w.background
 	fg := w.foreground
-	if !w.currentBackground.IsEmpty() {
-		bg = w.currentBackground
+	if !w.current.Background().IsEmpty() {
+		bg = w.current.Background()
 	}
-	if !w.currentForeground.IsEmpty() {
-		fg = w.currentForeground
+	if !w.current.Foreground().IsEmpty() {
+		fg = w.current.Foreground()
 	}
 
 	// ignore processing fully tranparent colors
@@ -404,7 +428,7 @@ func (w *Writer) writeSegmentColors() {
 		return
 	}
 
-	if fg.IsTransparent() && len(w.TerminalBackground) != 0 {
+	if fg.IsTransparent() && len(w.TerminalBackground) != 0 { //nolint: gocritic
 		background := w.getAnsiFromColorString(w.TerminalBackground, false)
 		w.writeEscapedAnsiString(fmt.Sprintf(colorise, background))
 		w.writeEscapedAnsiString(fmt.Sprintf(colorise, bg.ToForeground()))
@@ -421,45 +445,14 @@ func (w *Writer) writeSegmentColors() {
 	}
 
 	// set current colors
-	w.currentBackground = bg
-	w.currentForeground = fg
+	w.current.Add(bg, fg)
 }
 
-func (w *Writer) writeColorOverrides(match map[string]string, background string, i int) int {
+func (w *Writer) writeArchorOverride(match map[string]string, background string, i int) int {
 	position := i
 	// check color reset first
 	if match[ANCHOR] == resetStyle.AnchorEnd {
-		// make sure to reset the colors if needed
-		position += len([]rune(resetStyle.AnchorEnd)) - 1
-
-		// do not reset when colors are identical
-		if w.currentBackground == w.background && w.currentForeground == w.foreground {
-			return position
-		}
-
-		// do not restore colors at the end of the string, we print it anyways
-		if position == len(w.runes)-1 {
-			return position
-		}
-
-		if w.transparent {
-			w.writeEscapedAnsiString(transparentEnd)
-		}
-
-		if w.background.IsClear() {
-			w.writeEscapedAnsiString(backgroundStyle.End)
-		}
-
-		if w.currentBackground != w.background && !w.background.IsClear() {
-			w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.background))
-		}
-
-		if (w.currentForeground != w.foreground || w.transparent) && !w.foreground.IsClear() {
-			w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.foreground))
-		}
-
-		w.transparent = false
-		return position
+		return w.endColorOverride(position)
 	}
 
 	position += len([]rune(match[ANCHOR])) - 1
@@ -478,51 +471,119 @@ func (w *Writer) writeColorOverrides(match map[string]string, background string,
 	if match[FG] == Transparent && len(match[BG]) == 0 {
 		match[BG] = background
 	}
-	w.currentBackground, w.currentForeground = w.asAnsiColors(match[BG], match[FG])
+
+	bg, fg := w.asAnsiColors(match[BG], match[FG])
 
 	// ignore processing fully tranparent colors
-	w.invisible = w.currentForeground.IsTransparent() && w.currentBackground.IsTransparent()
+	w.invisible = fg.IsTransparent() && bg.IsTransparent()
 	if w.invisible {
 		return position
 	}
 
 	// make sure we have colors
-	if w.currentForeground.IsEmpty() {
-		w.currentForeground = w.foreground
+	if fg.IsEmpty() {
+		fg = w.foreground
 	}
-	if w.currentBackground.IsEmpty() {
-		w.currentBackground = w.background
+	if bg.IsEmpty() {
+		bg = w.background
 	}
 
-	if w.currentForeground.IsTransparent() && len(w.TerminalBackground) != 0 {
+	w.current.Add(bg, fg)
+
+	if w.current.Foreground().IsTransparent() && len(w.TerminalBackground) != 0 {
 		background := w.getAnsiFromColorString(w.TerminalBackground, false)
 		w.writeEscapedAnsiString(fmt.Sprintf(colorise, background))
-		w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.currentBackground.ToForeground()))
+		w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.current.Background().ToForeground()))
 		return position
 	}
 
-	if w.currentForeground.IsTransparent() && !w.currentBackground.IsTransparent() {
+	if w.current.Foreground().IsTransparent() && !w.current.Background().IsTransparent() {
 		w.transparent = true
-		w.writeEscapedAnsiString(fmt.Sprintf(transparent, w.currentBackground))
+		w.writeEscapedAnsiString(fmt.Sprintf(transparent, w.current.Background()))
 		return position
 	}
 
-	if w.currentBackground != w.background {
+	if w.current.Background() != w.background {
 		// end the colors in case we have a transparent background
-		if w.currentBackground.IsTransparent() {
+		if w.current.Background().IsTransparent() {
 			w.writeEscapedAnsiString(backgroundEnd)
 		} else {
-			w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.currentBackground))
+			w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.current.Background()))
 		}
 	}
 
-	if w.currentForeground != w.foreground {
-		w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.currentForeground))
+	if w.current.Foreground() != w.foreground {
+		w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.current.Foreground()))
 	}
 
 	return position
 }
 
+func (w *Writer) endColorOverride(position int) int {
+	// make sure to reset the colors if needed
+	position += len([]rune(resetStyle.AnchorEnd)) - 1
+
+	// do not restore colors at the end of the string, we print it anyways
+	if position == len(w.runes)-1 {
+		w.current.Pop()
+		return position
+	}
+
+	// reset colors to previous when we have more than 1 in stack
+	// as soon as we have  more than 1, we can pop the last one
+	// and print the previous override as it wasn't ended yet
+	if w.current.Len() > 1 {
+		fg := w.current.Foreground()
+		bg := w.current.Background()
+
+		w.current.Pop()
+
+		previousBg := w.current.Background()
+		previousFg := w.current.Foreground()
+
+		if w.transparent {
+			w.writeEscapedAnsiString(transparentEnd)
+		}
+
+		if previousBg != bg {
+			w.writeEscapedAnsiString(fmt.Sprintf(colorise, previousBg))
+		}
+
+		if previousFg != fg {
+			w.writeEscapedAnsiString(fmt.Sprintf(colorise, previousFg))
+		}
+
+		return position
+	}
+
+	// pop the last colors from the stack
+	defer w.current.Pop()
+
+	// do not reset when colors are identical
+	if w.current.Background() == w.background && w.current.Foreground() == w.foreground {
+		return position
+	}
+
+	if w.transparent {
+		w.writeEscapedAnsiString(transparentEnd)
+	}
+
+	if w.background.IsClear() {
+		w.writeEscapedAnsiString(backgroundStyle.End)
+	}
+
+	if w.current.Background() != w.background && !w.background.IsClear() {
+		w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.background))
+	}
+
+	if (w.current.Foreground() != w.foreground || w.transparent) && !w.foreground.IsClear() {
+		w.writeEscapedAnsiString(fmt.Sprintf(colorise, w.foreground))
+	}
+
+	w.transparent = false
+	return position
+}
+
 func (w *Writer) asAnsiColors(background, foreground string) (Color, Color) {
 	if len(background) == 0 {
 		background = Background
